'use strict'

const picomatch = require('picomatch')
const Opal = global.Opal

module.exports.register = function (registry, { file, contentCatalog }) {
  registry.blockMacro('indexList', function () {
    const self = this
    self.$option('format', 'short')
    self.process((parent, target, attrs) => {
      if (typeof attrs === 'object' && '$$smap' in attrs) {
        attrs = fromHash(attrs)
      }
      // console.log('attrs', attrs)
      if (attrs.level && (parent.blocks[parent.blocks.length - 1]).context === 'ulist') {
        parent = parent.blocks[parent.blocks.length - 1]
      }
      const list = computeParent(parent, attrs.level || 1, self)
      const marker = '*'.repeat(attrs.level || 1)

      doFiles(attrs, (item) => {
        // console.log(`list item`, item.src)
        const listItem = self.createListItem(list, xref(item))
        listItem.marker = marker
        list.blocks.push(listItem)
      })
    })
  })

  registry.blockMacro('indexTable', function () {
    const self = this
    self.$option('format', 'short')
    self.process((parent, target, attrs) => {
      if (typeof attrs === 'object' && '$$smap' in attrs) {
        attrs = fromHash(attrs)
      }
      // console.log('attrs', attrs)

      const table = parent.blocks[parent.blocks.length - 1]
      if (table.context !== 'table') {
        console.log(`indexTable block macro must follow a table, not ${table.context}`)
        return
      }
      const columns = table.columns
      const rows = table.rows.body
      const cellAttrs = attrs.cells ? attrs.cells.split(',') : []
      if (columns.length !== cellAttrs.length) {
        console.log(`column count ${columns.length} differs from 'cells' attribute count ${cellAttrs.length}`)
        cellAttrs.length = columns.length
      }

      doFiles(attrs, (item) => {
        // console.log(`list item`, item.src)
        const attributes = item[1]
        // console.log('attributes', attributes)
        attributes.$xref = xref(item)
        const row = columns.map((col, i) =>
          Opal.Asciidoctor.Table.Cell.$new(col, attributes[cellAttrs[i]] || ''))
        rows.push(row)
      })
    })
  })

  registry.inlineMacro('indexCount', function () {
    const self = this
    self.$option('format', 'short')
    self.process((parent, target, attrs) => {
      if (typeof attrs === 'object' && '$$smap' in attrs) {
        attrs = fromHash(attrs)
      }
      var count = 0
      doFiles(attrs, (item) => count++)
      return self.createInline(parent, 'quoted', count)
    })
  })

  registry.inlineMacro('indexUniqueCount', function () {
    const self = this
    self.$option('format', 'short')
    self.process((parent, target, attrs) => {
      if (typeof attrs === 'object' && '$$smap' in attrs) {
        attrs = fromHash(attrs)
      }
      const unique = attrs.unique
      if (!unique) {
        console.log('attribute \'unique\' must be specified')
        return 'attribute \'unique\' must be specified'
      }
      var values = new Set()
      doFiles(attrs, (item) => values.add(item[1][unique]))
      return self.createInline(parent, 'quoted', values.size)
    })
  })

  function doFiles (attrs, content) {
    const { filter, relCompare, attributesPresent, attributesMissing, attributeValues } = getFilter(attrs)
    const files = contentCatalog.findBy(filter)
      .filter((test) => (test.src.family === 'page' || test.src.family === 'alias') &&
        test !== file &&
        test.src.component && //never include site start page alias
        relCompare(test.src.relative))
      .map((item) => [item, attributes(item)])
      .filter((item) =>
        attributesPresent.every((attributeName) => attributeName in item[1]) &&
        attributesMissing.every((attributeName) => !(attributeName in item[1])) &&
        attributeValues.every(([name, value]) => item[1][name] === value)
      )
    files.sort(compare)

    files.forEach(content)
  }

  function getFilter (attrs) {
    const { component, version, module } = file.src
    const filter = {}
    attrs.component === '*' || (filter.component = attrs.component || component)
    attrs.version === '*' || (filter.version = attrs.version || version)
    attrs.module === '*' || (filter.module = attrs.module || module)
    const relCompare = attrs.relative ? picomatch(attrs.relative) : () => true
    const attributesPresent = []
    const attributesMissing = []
    const attributeValues = []
    attrs.attributes && attrs.attributes.split(',').forEach((clause) => {
      const match = clause.split('=')
      if (match.length === 2) {
        attributeValues.push([match[0], match[1]])
      } else if (match.length === 1) {
        if (match[0].startsWith('!')) {
          attributesMissing.push(match[0].slice(1))
        } else {
          attributesPresent.push(match[0])
        }
      } else {
        console.log(`invalid clause ${clause}`)
      }
    })
    return { filter, relCompare, attributesPresent, attributesMissing, attributeValues }
  }
}

const compare = (f1, f2) => {
  const t1 = f1[1].doctitle.toLowerCase()
  const t2 = f2[1].doctitle.toLowerCase()
  return t1 < t2 ? -1 : t1 > t2 ? 1 : 0
}

function xref (item) {
  const { component, version, module, relative } = item[0].src
  return `xref:${version}@${component}:${module}:${relative}[${item[1].doctitle}]`
}

function attributes (file) {
  file = file.src.family === 'page' ? file : file.rel
  return Object.assign({ doctitle: file.src.relative }, file.asciidoc ? file.asciidoc.attributes : {})
}

function computeParent (parent, depth, self) {
  // console.log(`depth ${depth}, parent.context`, parent.context)
  if (parent.context === 'ulist') {
    if (depth <= 1) {
      return parent
    }
    if (parent.blocks.length) {
      const last = parent.blocks[parent.blocks.length - 1]
      //last is a list_item
      if (last.blocks.length) {
        const lastBlock = last.blocks[last.blocks.length - 1]
        if (lastBlock.context === 'ulist') {
          return computeParent(lastBlock, depth - 1, self)
        }
      }
      // console.log('about to add to existing list item')
      return computeParent(last, depth, self)
    }
    //list with no items.... how did this happen?
    // console.log('new list item')
    const listItem = self.createListItem(parent, 'How did this happen?')
    parent.blocks.push(listItem)
    return computeParent(listItem, depth, self)
  }
  // console.log('new ulist')
  const list = self.createList(parent, 'ulist')
  parent.blocks.push(list)
  return computeParent(list, depth - 1, self)
}

function fromHash (hash) {
  const object = {}
  const data = hash.$$smap
  for (const key in data) {
    object[key] = data[key]
  }
  return object
}

// function print (block, level = 0) {
//   console.log(`level: ${level}`, block)
//   level++
//   block.blocks.forEach((sub) => print(sub, level))
// }
